#define LOCK_PROTOTYPE CK_RWCOHORT_NEUTRAL_PROTOTYPE
#define LOCK_INSTANCE CK_RWCOHORT_NEUTRAL_INSTANCE
#define LOCK_INITIALIZER CK_RWCOHORT_NEUTRAL_INITIALIZER
#define LOCK_INIT(N, C, W) CK_RWCOHORT_NEUTRAL_INIT(N, C)
#define READ_LOCK CK_RWCOHORT_NEUTRAL_READ_LOCK
#define WRITE_LOCK CK_RWCOHORT_NEUTRAL_WRITE_LOCK
#define READ_UNLOCK CK_RWCOHORT_NEUTRAL_READ_UNLOCK
#define WRITE_UNLOCK CK_RWCOHORT_NEUTRAL_WRITE_UNLOCK
