#define LOCK_PROTOTYPE CK_RWCOHORT_WP_PROTOTYPE
#define LOCK_INSTANCE CK_RWCOHORT_WP_INSTANCE
#define LOCK_INITIALIZER CK_RWCOHORT_WP_INITIALIZER
#define LOCK_INIT CK_RWCOHORT_WP_INIT
#define READ_LOCK CK_RWCOHORT_WP_READ_LOCK
#define WRITE_LOCK CK_RWCOHORT_WP_WRITE_LOCK
#define READ_UNLOCK CK_RWCOHORT_WP_READ_UNLOCK
#define WRITE_UNLOCK CK_RWCOHORT_WP_WRITE_UNLOCK
